import os
import sys

from sqlalchemy import engine_from_config

from pyramid.paster import (
    get_appsettings,
    setup_logging,
    )

from pyramid.scripts.common import parse_vars

from ..models import Base
from ..settings import apply_local_settings


def usage(argv):
    cmd = os.path.basename(argv[0])
    print('usage: %s <config_uri> [var=value]\n'
          '(example: "%s development.ini")' % (cmd, cmd))
    sys.exit(1)


def main(argv=sys.argv):
    """
    Function called when the "initialize_c2cgeoform_db" script is run. It
    creates in the database the tables declared in the model of the test
    application ("pully").
    """
    if len(argv) < 2:
        usage(argv)
    config_uri = argv[1]
    options = parse_vars(argv[2:])
    setup_logging(config_uri)
    settings = get_appsettings(config_uri, options=options)
    apply_local_settings(settings)
    engine = engine_from_config(settings, 'sqlalchemy.')

    # FIXME this is needed for now so the "Pully" model is in the
    # metadata object when create_all is called.
    from ..pully.model import ExcavationPermission  # flake8: noqa
    
    # Base.metadata.drop_all(engine)
    Base.metadata.create_all(engine)



# coding=utf-8
from sqlalchemy import (
    Column,
    Integer,
    BigInteger,
    Text,
    Boolean,
    Date,
    ForeignKey)
from sqlalchemy.orm import relationship
import geoalchemy2

import colander
import deform
from translationstring import TranslationStringFactory

from c2cgeoform.ext import colander_ext, deform_ext
from c2cgeoform.models import Base, FileData

_ = TranslationStringFactory('pully')


# FIXME a file upload memory store is not appropriate for production
# See http://docs.pylonsproject.org/projects/deform/en/latest/interfaces.html#deform.interfaces.FileUploadTempStore  # noqa
class FileUploadTempStore(dict):
    def preview_url(self, name):
        return None

_file_upload_temp_store = FileUploadTempStore()


class District(Base):
    __tablename__ = 'district'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False)


class Address(Base):
    __tablename__ = 'address'

    id = Column(Integer, primary_key=True)
    label = Column(Text, nullable=False)


class ContactPerson(Base):
    __tablename__ = 'contact_person'
    __colanderalchemy_config__ = {
        'title': _('Contact Person')
    }

    id = Column(Integer, primary_key=True, info={
        'colanderalchemy': {
            'widget': deform.widget.HiddenWidget()
        }})
    first_name = Column(Text, nullable=False, info={
        'colanderalchemy': {
            'title': _('First name')
        }})
    last_name = Column(Text, nullable=False, info={
        'colanderalchemy': {
            'title': _('Last name')
        }})
    permission_id = Column(Integer, ForeignKey('excavations.id'), info={
        'colanderalchemy': {
            'widget': deform.widget.HiddenWidget()
        }})
    verified = Column(Boolean)


class Photo(FileData, Base):
    __tablename__ = 'photo'
    # Setting unknown to 'preserve' is required in classes used as a
    # FileUpload field.
    __colanderalchemy_config__ = {
        'title': _('Photo'),
        'unknown': 'preserve',
        'widget': deform_ext.FileUploadWidget(_file_upload_temp_store)
    }
    permission_id = Column(Integer, ForeignKey('excavations.id'))


class Situation(Base):
    __tablename__ = 'situation'

    id = Column(Integer, primary_key=True)
    name = Column(Text, nullable=False)
    name_fr = Column(Text, nullable=False)


class SituationForPermission(Base):
    __tablename__ = 'situation_for_permission'

    id = Column(Integer, primary_key=True)
    situation_id = Column(
        Integer, ForeignKey('situation.id'))
    permission_id = Column(
        Integer, ForeignKey('excavations.id'))


# This is the main model class which is used to register a schema.
class ExcavationPermission(Base):
    __tablename__ = 'excavations'
    __colanderalchemy_config__ = {
        'title':
        _('Application form for permission to carry out excavation work')
    }

    id = Column(Integer, primary_key=True, info={
        # the `colanderalchemy` property allows to set a custom title for the
        # column or to use a specific widget.
        'colanderalchemy': {
            'title': _('Permission Number'),
            'widget': deform.widget.HiddenWidget(),
            # if the `admin_list` property is enabled for a column on the main
            # model, then this column will be shown in the admin list grid.
            'admin_list': True
        }})
    # the hash column is required for all main models
    hash = Column(Text, unique=True)
    reference_number = Column(Text, nullable=True, info={
        'colanderalchemy': {
            'title': _('Reference Number')
        }})
    request_date = Column(Date, nullable=True, info={
        'colanderalchemy': {
            'title': _('Request Date'),
            'admin_list': True
        }})

    description = Column(Text, nullable=True, info={
        'colanderalchemy': {
            'title': _('Description of the Work'),
            'widget': deform.widget.TextAreaWidget(rows=3),
        }})
    motif = Column(Text, nullable=True, info={
        'colanderalchemy': {
            'title': _('Motive for the Work'),
            'widget': deform.widget.TextAreaWidget(rows=3),
        }})
    situations = relationship(
        SituationForPermission,
        # by setting the `cascade` property to 'delete-orphan', situation
        # entities are deleted when they are removed from the relation.
        cascade="all, delete-orphan",
        info={
            'colanderalchemy': {
                'title': _('Situation'),
                # this widget type shows a select widget where the values are
                # loaded from a database table. in this case the select options
                # are generated from the Situation table.
                'widget': deform_ext.RelationSelect2Widget(
                    Situation,
                    'id',
                    'name',
                    order_by='name',
                    multiple=True
                )
            }})
    # by default a Deform sequence widget is used for relationship columns,
    # which, for example, allows to create new contact persons in a sub-form.
    contact_persons = relationship(
        ContactPerson,
        # make sure persons are deleted when removed from the relation
        cascade="all, delete-orphan",
        info={
            'colanderalchemy': {
                'title': _('Contact Persons')
            }})
    location_district_id = Column(Integer, ForeignKey('district.id'), info={
        'colanderalchemy': {
            'title': _('District'),
            'widget': deform_ext.RelationSelect2Widget(
                District,
                'id',
                # if the name for the options should be internationalized, one
                # can create columns like 'name_fr' and 'name_de' in the table
                # 'District'. then in the translation files, the column name
                # can be "translated" (e.g. the French "translation" for the
                # column name would be 'name_fr'). to apply the translation use
                # the label `_('name'))` instead of `name`.
                'name',
                order_by='name',
                default_value=('', _('- Select -')),
            )
        }})
    location_street = Column(Text, nullable=False, info={
        'colanderalchemy': {
            'title': _('Street')
        }})
    location_postal_code = Column(Text, nullable=False, info={
        'colanderalchemy': {
            'title': _('Postal Code')
        }})
    location_town = Column(Text, nullable=False, info={
        'colanderalchemy': {
            'title': _('Town')
        }})
    # this is a search field to search for an address
    address_id = Column(Integer, ForeignKey('address.id'), info={
        'colanderalchemy': {
            'title': _('Address'),
            'widget': deform_ext.RelationSearchWidget(
                url=lambda request: request.route_url('addresses'),
                model=Address,
                min_length=1,
                id_field='id',
                label_field='label'
            )
        }})
    # to show a map for a geometry column, the column has to be defined as
    # follows.
    location_position = Column(
        geoalchemy2.Geometry('POINT', 4326, management=True), info={
            'colanderalchemy': {
                'title': _('Position'),
                'typ':
                colander_ext.Geometry('POINT', srid=4326, map_srid=3857),
                'widget': deform_ext.MapWidget()
            }})

    # Person in Charge for the Work
    responsible_title = Column(Text, nullable=True, info={
        'colanderalchemy': {
            'title': _('Title'),
            'validator': colander.OneOf(['mr', 'mrs']),
            'widget': deform.widget.SelectWidget(values=(
                ('', _('- Select -')),
                ('mr', _('Mr.')),
                ('mrs', _('Mrs.'))
            ))
        }})
    responsible_name = Column(Text, nullable=True, info={
        'colanderalchemy': {
            'title': _('Name'),
            'admin_list': True
        }})
    responsible_first_name = Column(Text, nullable=True, info={
        'colanderalchemy': {
            'title': _('First Name'),
            'admin_list': True
        }})
    responsible_mobile = Column(Text, nullable=True, info={
        'colanderalchemy': {
            'title': _('Mobile Phone')
        }})
    responsible_mail = Column(Text, nullable=True, info={
        'colanderalchemy': {
            'title': _('Mail'),
            'validator': colander.Email()
        }})
    responsible_company = Column(Text, nullable=True, info={
        'colanderalchemy': {
            'title': _('Company')
        }})

    validated = Column(Boolean, info={
        'colanderalchemy': {
            'title': _('Validation'),
            'label': _('Validated')
        }})

    # Work footprint
    work_footprint = Column(
        geoalchemy2.Geometry('MULTIPOLYGON', 4326, management=True), info={
            'colanderalchemy': {
                'title': _('Footprint for the Work'),
                'typ': colander_ext.Geometry(
                    'MULTIPOLYGON', srid=4326, map_srid=3857),
                'widget': deform_ext.MapWidget()
            }})

    # Photo
    photos = relationship(
        Photo,
        cascade="all, delete-orphan",
        info={
            'colanderalchemy': {
                'title': _('Photo')
            }})


class BusStop(Base):
    __tablename__ = 'bus_stops'

    id = Column(BigInteger, primary_key=True)
    name = Column(Text)
    geom = Column(geoalchemy2.Geometry('POINT', 4326, management=True))


class Comment(Base):
    __tablename__ = 'comments'
    __colanderalchemy_config__ = {
        'title': 'A very simple form'
    }

    id = Column(Integer, primary_key=True, info={
        'colanderalchemy': {
            'widget': deform.widget.HiddenWidget(),
            'admin_list': True
        }})
    hash = Column(Text, unique=True)
    name = Column(Text, nullable=False, info={
        'colanderalchemy': {
            'title': 'Name'
        }})
    comment = Column(Text, nullable=True, info={
        'colanderalchemy': {
            'title': 'Comment',
            'widget': deform.widget.TextAreaWidget(rows=3),
        }})
    bus_stop_id = Column(BigInteger, ForeignKey('bus_stops.id'), info={
        'colanderalchemy': {
            'title': 'Bus stop',
            'widget': deform_ext.RelationSelectMapWidget(
                label_field='name',
                url=lambda request: request.route_url('bus_stops')
            )
        }})


def setup_test_data(settings):
    from c2cgeoform.models import DBSession
    import transaction

    if DBSession.query(District).get(0) is None:
        DBSession.add(District(id=0, name="Pully"))
        DBSession.add(District(id=1, name="Paudex"))
        DBSession.add(District(id=2, name="Belmont-sur-Lausanne"))
        DBSession.add(District(id=3, name="Trois-Chasseurs"))
        DBSession.add(District(id=4, name="La Claie-aux-Moines"))
        DBSession.add(District(id=5, name="Savigny"))
        DBSession.add(District(id=6, name="Mollie-Margot"))

    if DBSession.query(Situation).get(0) is None:
        DBSession.add(Situation(id=0, name="Road", name_fr="Route"))
        DBSession.add(Situation(id=1, name="Sidewalk", name_fr="Trottoir"))
        DBSession.add(Situation(id=2, name="Berm", name_fr="Berme"))
        DBSession.add(Situation(
            id=3, name="Vegetated berm", name_fr="Berme végétalisée"))
        DBSession.add(Situation(id=4, name="Green zone", name_fr="Zone verte"))
        DBSession.add(Situation(id=5, name="Cobblestone", name_fr="Pavés"))

    if DBSession.query(BusStop).count() == 0:
        _add_bus_stops(DBSession, settings['bus_stops_file'])

    if DBSession.query(Address).get(0) is None:
        DBSession.add(Address(id=0, label="Bern"))
        DBSession.add(Address(id=1, label="Lausanne"))
        DBSession.add(Address(id=2, label="Genève"))
        DBSession.add(Address(id=3, label="Zurich"))
        DBSession.add(Address(id=4, label="Lugano"))

    transaction.commit()


def _add_bus_stops(session, filepath):
    """
    Load test data from a GeoJSON file.
    """
    import json
    from shapely.geometry import shape

    file = open(filepath)
    geojson = json.load(file)
    file.close()

    bus_stops = []
    for feature in geojson['features']:
        id = feature['id'].replace('node/', '')
        geometry = shape(feature['geometry'])
        name = feature['properties']['name'] \
            if 'name' in feature['properties'] else ''
        bus_stop = BusStop(
            id=long(float(id)),
            geom='SRID=4326;' + geometry.wkt,
            name=name)
        bus_stops.append(bus_stop)

    session.add_all(bus_stops)



from pyramid.response import Response
from pyramid.httpexceptions import HTTPNotFound, HTTPFound, HTTPBadRequest
from deform import Form, ValidationFailure, ZPTRendererFactory
from deform.form import Button
from deform.widget import FormWidget
from sqlalchemy import desc, or_, types
from sqlalchemy.orm.exc import NoResultFound
from geoalchemy2.elements import WKBElement
from colander import SchemaNode, Mapping
from translationstring import TranslationStringFactory
import uuid

from .models import DBSession
from .schema import forms
from .ext.deform_ext import RecaptchaWidget

_ = TranslationStringFactory('c2cgeoform')


def _get_schema(request):
    schema_name = request.matchdict['schema']

    if _is_favicon_request(schema_name):
        # send a 404 for favicon requests that were mapped to the route
        raise HTTPNotFound()
    elif schema_name in forms:
        return forms.get(schema_name)
    else:
        raise HTTPNotFound('invalid schema \'' + schema_name + '\'')


def notfound(request):
    return HTTPNotFound()


def form(request):
    geo_form_schema = _get_schema(request)
    session = request.session
    with_captcha = \
        geo_form_schema.show_captcha and not geo_form_schema.show_confirmation
    form = _get_form(geo_form_schema, 'form', request, with_captcha)

    submission_id = request.params.get('__submission_id__')
    if len(request.POST) > 0 or submission_id is not None:
        # there is form data either in the session or as POST data

        if submission_id is not None and submission_id in session:
            # restoring the form from the session when going back from the
            # confirmation page
            form_data = session[submission_id]
            custom_data = None
            only_validate = '1'
        else:
            form_data = request.POST.items()
            custom_data = request.POST.get('__custom_data__')
            only_validate = request.POST.get('__only_validate__')

        try:
            obj_dict = form.validate(form_data)
        except ValidationFailure as e:
            # FIXME see https://github.com/Pylons/deform/pull/243
            rendered = e.field.widget.serialize(
                e.field, e.cstruct, custom_data=custom_data, request=request)
        else:
            if only_validate == '1':
                # even if the validation was successful, show the form again
                # (required for the wizzard)
                rendered = form.render(obj_dict, custom_data=custom_data,
                                       request=request)
            else:
                # the form data is valid, redirect to confirmation page.
                # store the form data in a session, so that it can be restored.
                submission_id = str(uuid.uuid4())
                session[submission_id] = form_data
                session.save()

                url = request.route_url(
                    'confirm', schema=geo_form_schema.name,
                    _query={'__submission_id__': submission_id})
                return HTTPFound(url)

    else:
        # empty form
        rendered = form.render(custom_data=None, request=request)

    return {'form': rendered,
            'deform_dependencies': form.get_widget_resources()}


def confirmation(request):
    geo_form_schema = _get_schema(request)
    session = request.session
    with_captcha = \
        geo_form_schema.show_captcha and geo_form_schema.show_confirmation
    form = _get_form(
        geo_form_schema, 'form_confirmation', request, False)

    submission_id = request.params.get('__submission_id__')
    if submission_id is None or submission_id not in session:
        raise HTTPBadRequest()

    store_form = request.params.get('__store_form__')
    form_data = session[submission_id]

    if geo_form_schema.show_confirmation:
        back_link = request.route_url(
            'form', schema=geo_form_schema.name,
            _query={'__submission_id__': submission_id})
        if store_form != '1':
            return _show_confirmation_page(
                geo_form_schema, form, form_data, with_captcha, submission_id,
                back_link, request)
        else:
            return _validate_and_store(
                geo_form_schema, form, form_data, with_captcha,
                back_link, request)
    else:
        # directly store without showing a confirmation page
        try:
            obj_dict = form.validate(form_data)
        except ValidationFailure:
            # something went wrong: the form data in the session is invalid
            raise HTTPBadRequest()
        return _store_in_db(geo_form_schema, obj_dict, request)


def _show_confirmation_page(
        geo_form_schema, form, form_data, with_captcha, submission_id,
        back_link, request):
    try:
        obj_dict = form.validate(form_data)
    except ValidationFailure:
        # something went wrong: the form data in the session is invalid
        raise HTTPBadRequest()
    if with_captcha:
        # get a form with a captcha widget
        form = _get_form(
            geo_form_schema, 'form_confirmation', request, True)

    rendered = form.render(
        obj_dict, readonly=True,
        custom_data=None, submission_id=submission_id,
        back_link=back_link)

    return {'form': rendered,
            'deform_dependencies': form.get_widget_resources()}


def _validate_and_store(
        geo_form_schema, form, form_data, with_captcha,
        back_link, request):
    if with_captcha:
        form = _get_form(
            geo_form_schema, 'form_confirmation', request, True)
        form_data = form_data + request.POST.items()
    try:
        obj_dict = form.validate(form_data)
    except ValidationFailure as e:
        # FIXME see https://github.com/Pylons/deform/pull/243
        rendered = e.field.widget.serialize(
            e.field, e.cstruct, readonly=True, request=request,
            back_link=back_link)
        return {'form': rendered,
                'deform_dependencies': form.get_widget_resources()}
    else:
        return _store_in_db(geo_form_schema, obj_dict, request)


def _store_in_db(geo_form_schema, obj_dict, request):
    # store in the database
    obj = geo_form_schema.schema_user.objectify(obj_dict)
    hash = str(uuid.uuid4())
    setattr(obj, geo_form_schema.hash_column_name, hash)
    DBSession.add(obj)

    url = request.route_url(
        'view_user',
        schema=geo_form_schema.name,
        hash=hash)

    return HTTPFound(url)


def _get_form(geo_form_schema, template, request, with_captcha=False):
    renderer = _get_renderer(geo_form_schema.templates_user)
    schema = geo_form_schema.schema_user
    if with_captcha:
        schema = _add_captcha(geo_form_schema, schema)
    form_action = request.route_url('form', schema=geo_form_schema.name)
    submit_button = Button(name='formsubmit', title=_('Submit'))

    form = Form(
        schema, buttons=(submit_button,),
        renderer=renderer, action=form_action)
    _set_form_widget(form, geo_form_schema.schema_user, template)
    _populate_widgets(form.schema, DBSession, request)

    return form


def _add_captcha(geo_form_schema, schema):
    schema = schema.clone()
    schema.add(SchemaNode(
        Mapping(),
        name='captcha',
        widget=RecaptchaWidget(
            public_key=geo_form_schema.recaptcha_public_key,
            private_key=geo_form_schema.recaptcha_private_key)))
    return schema


def _set_form_widget(form, schema, template):
    if getattr(schema, 'widget', None) is None:
        # only change if the user did not provide a custom widget
        form.widget = FormWidget(
            readonly_template='readonly/' + template, template=template)


def view_user(request):
    if 'hash' not in request.matchdict or not request.matchdict['hash']:
        return {'form': None}

    hash = request.matchdict['hash']
    geo_form_schema = _get_schema(request)
    renderer = _get_renderer(geo_form_schema.templates_user)
    form = Form(geo_form_schema.schema_user, renderer=renderer)
    _set_form_widget(form, geo_form_schema.schema_user, 'form_view_user')
    _populate_widgets(form.schema, DBSession, request)

    hash_field = getattr(
        geo_form_schema.model, geo_form_schema.hash_column_name)
    try:
        obj = DBSession \
            .query(geo_form_schema.model) \
            .filter(hash_field == hash) \
            .one()
    except NoResultFound:
        return {'form': None}

    rendered = form.render(geo_form_schema.schema_user.dictify(obj),
                           readonly=True, request=request)
    return {
        'form': rendered,
        'schema': geo_form_schema,
        'deform_dependencies': form.get_widget_resources()}


def list(request):
    geo_form_schema = _get_schema(request)
    entities = DBSession.query(geo_form_schema.model).all()
    return {'entities': entities, 'schema': geo_form_schema}


def grid(request):
    """API method which serves the JSON data for the Bootgrid table
    in the admin view.
    """
    geo_form_schema = _get_schema(request)

    current_page = int(request.POST.get('current'))
    row_count = int(request.POST.get('rowCount'))
    search_phrase = request.POST.get('searchPhrase', '').strip()
    sort = _get_sort_param(request.POST)

    query = _get_query(geo_form_schema, sort, search_phrase). \
        offset(current_page * row_count). \
        limit(row_count)

    return {
        "current": current_page,
        "rowCount": row_count,
        "rows": _get_grid_rows(query.all(), geo_form_schema),
        "total": query.count()
    }


def _get_sort_param(params):
    for key in params:
        # Bootgrid sends the sort field as "sort[first_name]: asc"
        if key.startswith('sort'):
            field = key.replace('sort[', '').replace(']', '')
            sort_order = 'asc' if params[key] == 'asc' else 'desc'
            return (field, sort_order)
    return None


def _get_query(geo_form_schema, sort, search_phrase):
    query = DBSession.query(geo_form_schema.model)

    # order by
    if sort is not None and hasattr(geo_form_schema.model, sort[0]):
        sort_field = getattr(geo_form_schema.model, sort[0])
        if sort[1] == 'desc':
            sort_field = desc(sort_field)
        query = query.order_by(sort_field)

    # search
    if search_phrase != '':
        search_expr = '%' + '%'.join(search_phrase.split()) + '%'

        # create `ilike` filters for every list text field
        filters = []
        for field in geo_form_schema.list_fields:
            column = getattr(geo_form_schema.model, field)
            # NOTE only text fields are searched
            if isinstance(column.type, types.String):
                like = getattr(column, 'ilike')
                filters.append(like(search_expr))

        # then join the filters into one `or` condition
        if len(filters) > 0:
            filter_expr = filters.pop()
            filter_expr = reduce(
                lambda filter_expr, filter: or_(filter_expr, filter),
                filters,
                filter_expr)
            query = query.filter(filter_expr)

    return query


def _get_grid_rows(entities, schema):
    """Creates plain objects for the given entities containing
    only those properties flagged with `admin_list`.
    """
    rows = []

    for entity in entities:
        obj = {}
        for field in schema.list_fields:
            value = getattr(entity, field)
            if value is None:
                value = ''
            else:
                if isinstance(value, WKBElement):
                    value = 'Geometry'
                else:
                    value = str(value)
            obj[field] = value
        # set the entity id on a special property
        obj['_id_'] = str(getattr(entity, schema.id_field))
        rows.append(obj)

    return rows


def edit(request):
    geo_form_schema = _get_schema(request)

    renderer = _get_renderer(geo_form_schema.templates_admin)
    form = Form(
        geo_form_schema.schema_admin, buttons=('submit',), renderer=renderer)
    _populate_widgets(form.schema, DBSession, request)

    if 'submit' in request.POST:
        form_data = request.POST.items()

        try:
            obj_dict = form.validate(form_data)
        except ValidationFailure as e:
            rendered = e.render()
        else:
            obj = geo_form_schema.schema_admin.objectify(obj_dict)
            obj = DBSession.merge(obj)
            DBSession.flush()

            # FIXME create a fresh form, otherwise the IDs of objects in
            # relationships will not be rendered
            # see https://github.com/Pylons/deform/issues/236
            form = Form(
                geo_form_schema.schema_admin, buttons=('submit',),
                renderer=renderer)

            rendered = form.render(geo_form_schema.schema_admin.dictify(obj),
                                   request=request)
    else:
        id = request.matchdict['id']
        obj = DBSession.query(geo_form_schema.model).get(id)
        rendered = form.render(geo_form_schema.schema_admin.dictify(obj),
                               request=request)

    return {
        'form': rendered,
        'schema': geo_form_schema,
        'deform_dependencies': form.get_widget_resources()}


def view_admin(request):
    id_ = request.matchdict['id']
    geo_form_schema = _get_schema(request)
    renderer = _get_renderer(geo_form_schema.templates_admin)
    form = Form(geo_form_schema.schema_admin, buttons=('submit',),
                renderer=renderer)
    obj = DBSession.query(geo_form_schema.model).get(id_)
    rendered = form.render(geo_form_schema.schema_admin.dictify(obj),
                           readonly=True, request=request)
    return {
        'form': rendered,
        'schema': geo_form_schema,
        'deform_dependencies': form.get_widget_resources()}


def set_locale_cookie(request):
    """ View to change the preferred language.
    """
    if request.GET['language']:
        language = request.GET['language']
        response = Response()
        response.set_cookie('_LOCALE_',
                            value=language,
                            max_age=31536000)  # max_age = year
    return HTTPFound(location=request.environ['HTTP_REFERER'],
                     headers=response.headers)


def _populate_widgets(node, session, request):
    """ Populate ``deform_ext.RelationSelectMixin`` widgets.
    """
    if hasattr(node.widget, 'populate'):
        node.widget.populate(session, request)

    for child in node:
        _populate_widgets(child, session, request)


def _get_renderer(search_paths):
    if search_paths is None:
        return None
    else:
        from c2cgeoform import translator
        return ZPTRendererFactory(search_paths, translator=translator)


def _is_favicon_request(text):
    return text in [
        'favicon.ico',
        'apple-touch-icon-precomposed.png',
        'apple-touch-icon.png']



from pyramid.httpexceptions import HTTPBadRequest
from c2cgeoform.models import DBSession
from ..model import Address


def addresses(request):
    if 'term' not in request.params:
        return HTTPBadRequest()
    term = '%%%s%%' % request.params['term']
    query = DBSession.query(Address).filter(Address.label.ilike(term))
    return [{'id': addr.id, 'label': addr.label} for addr in query]



from sqlalchemy.orm import (class_mapper, ColumnProperty, RelationshipProperty)
from sqlalchemy import inspect
from colanderalchemy import SQLAlchemySchemaNode
from . import default_search_paths


class GeoFormSchema():

    _COLANDERALCHEMY = 'colanderalchemy'
    # key to indicate that a field should be shown in the admin list
    _ADMIN_LIST = 'admin_list'

    def __init__(
            self, name, model,
            templates_user=None, templates_admin=None,
            overrides_user=None, overrides_admin=None,
            includes_user=None, includes_admin=None,
            excludes_user=None, excludes_admin=None,
            hash_column_name='hash', show_confirmation=True,
            show_captcha=False, recaptcha_public_key=None,
            recaptcha_private_key=None, **kw):
        self.name = name
        self.model = model
        self.hash_column_name = hash_column_name
        self.show_confirmation = show_confirmation
        self.show_captcha = show_captcha

        if includes_user is not None:
            excludes_user = None
        else:
            excludes_user = [] if excludes_user is None else excludes_user
            excludes_user = excludes_user + [hash_column_name]

        self.schema_user = SQLAlchemySchemaNode(
            self.model,
            overrides=overrides_user,
            includes=includes_user,
            excludes=excludes_user,
            **kw)
        self.schema_admin = SQLAlchemySchemaNode(
            self.model,
            overrides=overrides_admin,
            includes=includes_admin,
            excludes=excludes_admin,
            **kw)

        self.templates_user = default_search_paths
        if templates_user is not None:
            self.templates_user = (templates_user,) + self.templates_user

        self.templates_admin = default_search_paths
        if templates_admin is not None:
            self.templates_admin = (templates_admin,) + self.templates_admin

        meta_model = class_mapper(model)
        if len(meta_model.primary_key) != 1:
            raise RuntimeError(
                'Model ' + name + ' must have exactly ' +
                'one primary key column')
        self.id_field = meta_model.primary_key[0].name

        if not self._has_hash_column():
            raise RuntimeError(
                'Model ' + name + ' does not contain a ' +
                'hash-value column: ' + hash_column_name)

        self.list_fields = self._get_fields_with_property(self._ADMIN_LIST)

        if show_captcha and (
                recaptcha_public_key is None or
                recaptcha_private_key is None):
            raise RuntimeError(
                '`recaptcha_public_key` and `recaptcha_private_key` must be ' +
                'set when using a captcha')
        self.recaptcha_public_key = recaptcha_public_key
        self.recaptcha_private_key = recaptcha_private_key

    def _get_fields_with_property(self, property):
        """ Search the columns where the given property is set to True.
        """
        fields = []
        mapper = inspect(self.model)
        for column in mapper.attrs:
            info = {}
            if isinstance(column, ColumnProperty):
                info = column.columns[0].info
            elif isinstance(column, RelationshipProperty):
                info = column.info

            if self._COLANDERALCHEMY in info and \
                    info[self._COLANDERALCHEMY].get(property, False):
                fields.append(column.key)
        return fields

    def _has_hash_column(self):
        """ Checks if the model contains the hash column.
        """
        mapper = inspect(self.model)
        for column in mapper.attrs:
            if isinstance(column, ColumnProperty):
                if column.columns[0].name == self.hash_column_name:
                    return True
        return False


forms = {}


def register_schema(
        name, model,
        templates_user=None, templates_admin=None,
        overrides_user=None, overrides_admin=None,
        includes_user=None, includes_admin=None,
        excludes_user=None, excludes_admin=None,
        hash_column_name='hash', show_confirmation=True, show_captcha=False,
        recaptcha_public_key=None, recaptcha_private_key=None):
    schema = GeoFormSchema(
        name, model, templates_user, templates_admin,
        overrides_user, overrides_admin, includes_user, includes_admin,
        excludes_user, excludes_admin, hash_column_name, show_confirmation,
        show_captcha, recaptcha_public_key, recaptcha_private_key)
    forms[name] = schema



from pyramid.config import Configurator
from sqlalchemy import engine_from_config
from pkg_resources import resource_filename
from pyramid.i18n import get_localizer
from pyramid.threadlocal import get_current_request
from pyramid.events import BeforeRender, NewRequest

from .models import DBSession
from .settings import apply_local_settings
from .subscribers import add_renderer_globals, add_localizer


def add_routes_and_views(config):
    # FIXME the append_slash stuff does not work as expected. When
    # using /{schema}/form instead /{schema}/form/ the route pattern
    # /{schema/{id} ("view_admin") will match and cause an exception.
    config.add_notfound_view('c2cgeoform.views.notfound',
                             append_slash=True)
    config.add_route('form', '/{schema}/form/')
    config.add_view('c2cgeoform.views.form',
                    route_name='form',
                    renderer='c2cgeoform:templates/site/form.pt')
    config.add_route('confirm', '/{schema}/form/confirm')
    config.add_view('c2cgeoform.views.confirmation',
                    route_name='confirm',
                    renderer='c2cgeoform:templates/site/confirmation.pt')
    config.add_route('view_user', '/{schema}/form/{hash}')
    config.add_view('c2cgeoform.views.view_user',
                    route_name='view_user',
                    renderer='c2cgeoform:templates/site/view_user.pt')
    config.add_route('list', '/{schema}/')
    config.add_view('c2cgeoform.views.list',
                    route_name='list',
                    renderer='c2cgeoform:templates/site/list.pt')
    config.add_route('grid', '/{schema}/grid/')
    config.add_view('c2cgeoform.views.grid',
                    route_name='grid', renderer='json',
                    request_method='POST')
    config.add_route('edit', '/{schema}/{id}/form')
    config.add_view('c2cgeoform.views.edit',
                    route_name='edit',
                    renderer='c2cgeoform:templates/site/edit.pt')
    config.add_route('view_admin', '/{schema}/{id}')
    config.add_view('c2cgeoform.views.view_admin',
                    route_name='view_admin',
                    renderer='c2cgeoform:templates/site/view_admin.pt')


def includeme(config):
    """
    Function called when "c2cgeoform" is included in a project (with
    ``config.include('c2cgeoform')``).

    This function creates routes and views for c2cgeoform pages.
    """
    config.include('pyramid_chameleon')
    config.include('pyramid_beaker')  # use Beaker for session storage
    config.add_static_view('c2cgeoform_static', 'static', cache_max_age=3600)
    config.add_static_view('deform_static', 'deform:static')
    config.add_route('locale', '/locale/')
    config.add_view('c2cgeoform.views.set_locale_cookie',
                    route_name='locale')
    config.add_translation_dirs('colander:locale', 'deform:locale', 'locale')
    config.add_directive('add_c2cgeoform_views', add_routes_and_views)
    _set_widget_template_path()

    config.add_subscriber(add_renderer_globals, BeforeRender)
    config.add_subscriber(add_localizer, NewRequest)


""" Default search paths for the form templates.
"""
default_search_paths = (
    resource_filename('c2cgeoform', 'templates/widgets'),
    resource_filename('deform', 'templates'))


def translator(term):
    request = get_current_request()
    if request is None:
        return term
    else:
        return get_localizer(request).translate(term)


def _set_widget_template_path():
    from deform import (Form, widget)

    Form.set_zpt_renderer(default_search_paths, translator=translator)

    registry = widget.ResourceRegistry()
    registry.set_js_resources(
        'json2', None, 'c2cgeoform:static/js/json2.min.js')
    registry.set_js_resources(
        'openlayers', '3.0.0', 'c2cgeoform:static/js/ol.js')
    registry.set_css_resources(
        'openlayers', '3.0.0', 'c2cgeoform:static/js/ol.css')
    registry.set_js_resources(
        'c2cgeoform.deform_map', None,
        'c2cgeoform:static/deform_map/controls.js')
    registry.set_css_resources(
        'c2cgeoform.deform_map', None,
        'c2cgeoform:static/deform_map/style.css')
    registry.set_js_resources(
        'typeahead', '0.10.5',
        'c2cgeoform:static/js/typeahead.bundle-0.10.5.min.js')
    registry.set_css_resources(
        'typeahead', '0.10.5',
        'c2cgeoform:static/js/typeaheadjs.css')
    registry.set_js_resources(
        'c2cgeoform.deform_search', None,
        'c2cgeoform:static/deform_search/search.js')
    Form.set_default_resource_registry(registry)


def main(global_config, **settings):
    """ This function returns a Pyramid WSGI application.

    This a is test application for the model and templates defined in
    c2cgeoform/pully.
    """
    apply_local_settings(settings)
    engine = engine_from_config(settings, 'sqlalchemy.')
    DBSession.configure(bind=engine)
    config = Configurator(settings=settings)
    config.include('pyramid_mako')
    config.include(includeme)

    config.add_translation_dirs('pully/locale')

    from .schema import register_schema
    from pully import model
    register_schema(
        'fouille',
        model.ExcavationPermission,
        templates_user=resource_filename('c2cgeoform', 'pully/templates'),
        excludes_user=['reference_number', 'validated'],
        overrides_user={
            # override the title for a field in the user form
            'request_date': {'title': 'Date'},
            # do not show the 'verified' field of ContactPerson for the user
            'contact_persons': {'excludes': ['verified']}
        },
        show_captcha=True,
        recaptcha_public_key=settings.get('recaptcha_public_key'),
        recaptcha_private_key=settings.get('recaptcha_private_key'))
    model.setup_test_data(settings)
    register_schema(
        'comment', model.Comment, show_confirmation=False, show_captcha=True,
        recaptcha_public_key=settings.get('recaptcha_public_key'),
        recaptcha_private_key=settings.get('recaptcha_private_key'))

    config.add_route('bus_stops', '/bus_stops')
    config.add_view('c2cgeoform.pully.views.bus_stops.bus_stops',
                    route_name='bus_stops', renderer='json',
                    request_method='GET')

    config.add_route('addresses', '/addresses')
    config.add_view('c2cgeoform.pully.views.addresses.addresses',
                    route_name='addresses', renderer='json',
                    request_method='GET')

    config.scan('c2cgeoform.pully')
    config.add_c2cgeoform_views()

    return config.make_wsgi_app()
